# -*- coding: utf-8 -*-
"""
Test
"""
import os
import numpy as np
from elm import TEST_DATA_DIR
from elm.embed import ChunkAndEmbed
import elm.embed

os.environ["OPENAI_API_KEY"] = "dummy"

FP_PDF = os.path.join(TEST_DATA_DIR, 'GPT-4.pdf')
FP_TXT = os.path.join(TEST_DATA_DIR, 'gpt4.txt')

with open(FP_TXT, 'r', encoding='utf8') as f:
    TEXT = f.read()


class MockClass:
    """Dummy class to mock ChunkAndEmbed.call_api()"""

    @staticmethod
    def call(*args, **kwargs):  # pylint: disable=unused-argument
        """Mock for ChunkAndEmbed.call_api()"""
        embedding = np.random.uniform(0, 1, 10)
        response = {'data': [{'embedding': embedding}]}
        return response


def test_chunk_and_embed(mocker):
    """Simple text to embedding test

    Note that embedding api is mocked here and not actually tested.
    """
    mocker.patch.object(elm.embed.ChunkAndEmbed, "call_api", MockClass.call)
    ce0 = ChunkAndEmbed(TEXT, tokens_per_chunk=100)
    embeddings = ce0.run()
    assert len(embeddings) == len(ce0.text_chunks)

    ce1 = ChunkAndEmbed(TEXT, tokens_per_chunk=400)
    embeddings = ce1.run()
    assert len(ce1.text_chunks) < len(ce0.text_chunks)
    assert len(embeddings) == len(ce1.text_chunks)

    for emb in embeddings:
        assert isinstance(emb, np.ndarray)
        assert len(emb) == 10



# -*- coding: utf-8 -*-
"""
Test
"""
import os
import numpy as np
from elm import TEST_DATA_DIR
from elm.chunk import Chunker


os.environ["OPENAI_API_KEY"] = "dummy"

FP_TXT = os.path.join(TEST_DATA_DIR, 'gpt4.txt')
with open(FP_TXT, 'r', encoding='utf8') as f:
    TEXT = f.read()


def test_overlap():
    """Test overlap of chunk paragraphs"""
    chunks0 = Chunker(TEXT, tokens_per_chunk=1, overlap=0)
    assert len(chunks0.text.split('\n\n')) == len(chunks0)

    chunks1 = Chunker(TEXT, tokens_per_chunk=1, overlap=1)
    assert len(chunks1.text.split('\n\n')) == len(chunks1)

    for i, c0 in enumerate(chunks0):
        assert len(c0) < len(chunks1[i])

    for i0 in [5, 10, 25]:
        c1 = chunks1[i0].split('\n\n')
        assert c1[0] == chunks0[i0 - 1]
        assert c1[1] == chunks0[i0]
        assert c1[2] == chunks0[i0 + 1]


def test_by_tokens():
    """Test chunking of text by token count"""
    chunks0 = Chunker(TEXT, tokens_per_chunk=100, overlap=0)
    chunks1 = Chunker(TEXT, tokens_per_chunk=200, overlap=0)
    chunks2 = Chunker(TEXT, tokens_per_chunk=400, overlap=0)

    assert len(chunks1) < len(chunks0)
    assert len(chunks2) < len(chunks0)

    assert np.max(chunks0.chunk_tokens) < 200
    assert np.max(chunks1.chunk_tokens) < 200
    assert np.max(chunks2.chunk_tokens) < 400

    assert sum(chunks0.chunk_tokens) == sum(chunks1.chunk_tokens)
    assert sum(chunks0.chunk_tokens) == sum(chunks2.chunk_tokens)

    assert len('\n\n'.join(chunks0.chunks)) == len('\n\n'.join(chunks1.chunks))
    assert len('\n\n'.join(chunks0.chunks)) == len('\n\n'.join(chunks2.chunks))



# -*- coding: utf-8 -*-
"""
Test
"""
import os
import pandas as pd
import numpy as np
from elm import TEST_DATA_DIR
from elm.embed import ChunkAndEmbed
from elm.wizard import EnergyWizard
import elm.embed

os.environ["OPENAI_API_KEY"] = "dummy"

FP_PDF = os.path.join(TEST_DATA_DIR, 'GPT-4.pdf')
FP_TXT = os.path.join(TEST_DATA_DIR, 'gpt4.txt')

with open(FP_TXT, 'r', encoding='utf8') as f:
    TEXT = f.read()


class MockObject:
    """Dummy class for mocking api response objects"""


class MockClass:
    """Dummy class to mock various api calls"""

    @staticmethod
    def get_embedding(*args, **kwargs):  # pylint: disable=unused-argument
        """Mock for ChunkAndEmbed.call_api()"""
        embedding = np.random.uniform(0, 1, 10)
        return embedding

    @staticmethod
    def call(*args, **kwargs):  # pylint: disable=unused-argument
        """Mock for ChunkAndEmbed.call_api()"""
        embedding = np.random.uniform(0, 1, 10)
        response = {'data': [{'embedding': embedding}]}
        return response

    @staticmethod
    def create(*args, **kwargs):  # pylint: disable=unused-argument
        """Mock for openai.ChatCompletion.create()"""
        # pylint: disable=attribute-defined-outside-init
        response = MockObject()
        response.choices = [MockObject()]
        response.choices[0].message = MockObject()
        response.choices[0].message.content = 'hello!'
        return response


def make_corpus(mocker):
    """Make a text corpus with embeddings for the wizard."""
    mocker.patch.object(elm.embed.ChunkAndEmbed, "call_api", MockClass.call)
    mocker.patch.object(elm.wizard.EnergyWizard, "get_embedding",
                        MockClass.get_embedding)

    ce0 = ChunkAndEmbed(TEXT, tokens_per_chunk=400)
    embeddings = ce0.run()
    corpus = []
    for i, emb in enumerate(embeddings):
        corpus.append({'text': ce0.text_chunks[i], 'embedding': emb,
                       'ref': 'source0'})
    return corpus


def test_chunk_and_embed(mocker):
    """Simple text to embedding test

    Note that embedding api is mocked here and not actually tested.
    """
    corpus = make_corpus(mocker)
    wizard = EnergyWizard(pd.DataFrame(corpus), token_budget=1000,
                          ref_col='ref')

    mocker.patch.object(wizard._client.chat.completions, "create",
                        MockClass.create)

    question = 'What time is it?'
    out = wizard.chat(question, stream=False, print_references=True)
    response_message, query, references, performance = out
    assert response_message.startswith('hello!')
    assert query.startswith(EnergyWizard.MODEL_INSTRUCTION)
    assert query.endswith(question)
    assert 'source0' in references
    assert isinstance(performance, dict)


def test_convo_query(mocker):
    """Query with multiple messages

    Note that embedding api is mocked here and not actually tested.
    """

    corpus = make_corpus(mocker)
    wizard = EnergyWizard(pd.DataFrame(corpus), token_budget=1000,
                          ref_col='ref')

    mocker.patch.object(wizard._client.chat.completions, "create",
                        MockClass.create)

    question1 = 'What time is it?'
    question2 = 'How about now?'

    query = wizard.chat(question1, stream=False, convo=True,
                        print_references=True)[1]
    assert question1 in query
    assert question2 not in query
    assert len(wizard.messages) == 3

    query = wizard.chat(question2, stream=False, convo=True,
                        print_references=True)[1]
    assert question1 in query
    assert question2 in query
    assert len(wizard.messages) == 5

    wizard.clear()
    assert len(wizard.messages) == 1

    query = wizard.chat(question2, stream=False, convo=True,
                        print_references=True)[1]
    assert question1 not in query
    assert question2 in query
    assert len(wizard.messages) == 3



# -*- coding: utf-8 -*-
"""
Test
"""
import os
import networkx as nx
from elm import TEST_DATA_DIR
from elm.base import ApiBase
from elm.tree import DecisionTree
import elm.embed

os.environ["OPENAI_API_KEY"] = "dummy"

FP_PDF = os.path.join(TEST_DATA_DIR, 'GPT-4.pdf')
FP_TXT = os.path.join(TEST_DATA_DIR, 'gpt4.txt')

with open(FP_TXT, 'r', encoding='utf8') as f:
    TEXT = f.read()


class MockClass:
    """Dummy class to mock various api calls"""

    chat_messages = [{'role': 'role', 'content': 'test'}]

    @staticmethod
    def chat(prompt):
        """Mock for api.chat()"""
        return prompt


def test_chunk_and_embed(mocker):
    """Simple text to embedding test

    Note that embedding api is mocked here and not actually tested.
    """
    mocker.patch.object(elm.tree.DecisionTree, "api", MockClass)

    graph = nx.DiGraph(text='hello', name='Grant',
                       api=ApiBase(model='gpt-35-turbo'))

    response_dict = {}

    # pylint: disable=unused-argument
    def callback(response, graph, node_name):
        response_dict.update({node_name: response})

    graph.add_node('init', prompt='Say {text} to {name}')
    graph.add_edge('init', 'next', condition=lambda x: 'Grant' in x)
    graph.add_node('next', prompt='How are you?', callback=callback)

    tree = DecisionTree(graph)
    tree.run()

    assert 'init' in tree.history
    assert 'next' in tree.history
    assert isinstance(tree['next']['response'], str)
    assert isinstance(response_dict['next'], str)



"""
Test
"""
import os
import ast
import json
from io import BytesIO
import numpy as np
from elm import TEST_DATA_DIR
from elm.wizard import EnergyWizardPostgres


FP_REF_TXT = os.path.join(TEST_DATA_DIR, 'postgres_ref_output.txt')
FP_QUERY_TXT = os.path.join(TEST_DATA_DIR, 'postgres_query_db.txt')

with open(FP_REF_TXT, 'r', encoding='utf8') as f:
    REF_TEXT = f.read()

with open(FP_QUERY_TXT, 'r', encoding='utf8') as f:
    QUERY_TEXT = f.read()

QUERY_TUPLE = ast.literal_eval(QUERY_TEXT)
REF_TUPLE = ast.literal_eval(REF_TEXT)

os.environ["EWIZ_DB_USER"] = "user"
os.environ["EWIZ_DB_PASSWORD"] = "password"


class Cursor:
    """Dummy class for mocking database cursor objects"""
    def __init__(self):
        self.query = None

    def execute(self, query, vars):  # pylint: disable=unused-argument
        """Mock for cursor.execute()
        Parameters:
        -----------
        query : str
            SQL query to send to database
        vars : str
            Variables to include in query, ex: embedding, ids
        """
        self.query = query

    def fetchall(self):
        """Mock for cursor.fetchall()"""
        if "vector as score" in self.query:
            return QUERY_TUPLE
        if "id IN (" in self.query:
            return REF_TUPLE
        else:
            return None


class BotoClient:
    """Dummy class for mocking boto3 objects"""

    def invoke_model(self, **kwargs):  # pylint: disable=unused-argument
        """Mock for boto3.invoke_model()"""
        dummy_json_content = json.dumps({"key": "value"}).encode('utf-8')

        dummy_response = {
            'statusCode': 200,
            'body': BytesIO(dummy_json_content)
        }

        return dummy_response


def test_postgres(mocker):
    """Test to ensure correct response vector db."""

    mock_conn_cm = mocker.MagicMock()
    mock_conn = mock_conn_cm.__enter__.return_value
    mock_conn.cursor.return_value = Cursor()

    mock_connect = mocker.patch('psycopg2.connect')
    mock_connect.return_value = mock_conn_cm
    wizard = EnergyWizardPostgres(db_host='Dummy', db_port='Dummy',
                                  db_name='Dummy', db_schema='Dummy',
                                  db_table='Dummy',
                                  boto_client=BotoClient())

    question = 'Is this a dummy question?'

    strings, scores, ids = wizard.query_vector_db(question)

    assert ids is not None
    assert len(strings) == len(scores)
    assert len(ids) == len(strings)

    ref_list = wizard.make_ref_list(ids)

    assert len(ref_list) > 0
    assert 'title' in str(ref_list)
    assert 'url' in str(ref_list)
    assert 'research-hub.nrel.gov' in str(ref_list)


def test_ref_replace():
    """Test to ensure removal of double quotes from references."""

    wizard = EnergyWizardPostgres(db_host='Dummy', db_port='Dummy',
                                  db_name='Dummy', db_schema='Dummy',
                                  db_table='Dummy', cursor=Cursor(),
                                  boto_client=BotoClient(),
                                  meta_columns=['title', 'url',
                                                'nrel_id', 'id'])

    refs = [(chr(34), 'test.com', 'nrel-1', '5a'),
            ('remove "double" quotes', 'test_2.com', 'nrel-2', '7b')]

    ids = np.array(['7b', '5a'])

    out = wizard._format_refs(refs, ids)

    assert len(out) > 1

    for i in out:
        refs_dict = json.loads(i)
        assert '"' not in refs_dict['title']
        assert chr(34) not in refs_dict['title']


def test_ids():
    """Test to ensure only records with valid ids are returned."""

    wizard = EnergyWizardPostgres(db_host='Dummy', db_port='Dummy',
                                  db_name='Dummy', db_schema='Dummy',
                                  db_table='Dummy', cursor=Cursor(),
                                  boto_client=BotoClient(),
                                  meta_columns=['title', 'url',
                                                'nrel_id', 'id'])

    refs = [('title', 'test.com', 'nrel-1', '5a'),
            ('title2', 'test_2.com', 'nrel-2', '7b')]

    ids = np.array(['7c', '5a'])

    out = wizard._format_refs(refs, ids)

    assert len(out) == 1
    assert not any('7b' in item for item in out)


def test_sorted_refs():
    """Test to ensure references are sorted in same order as ids."""

    wizard = EnergyWizardPostgres(db_host='Dummy', db_port='Dummy',
                                  db_name='Dummy', db_schema='Dummy',
                                  db_table='Dummy', cursor=Cursor(),
                                  boto_client=BotoClient(),
                                  meta_columns=['title', 'url',
                                                'nrel_id', 'id'])

    refs = [('title', 'test.com', 'nrel-1', '5a'),
            ('title2', 'test_2.com', 'nrel-2', '7b')]

    ids = np.array(['7b', '5a'])

    expected = [('{"title": "title2", "url": "test_2.com", '
                 '"nrel_id": "nrel-2", "id": "7b"}'),
                ('{"title": "title", "url": "test.com", '
                 '"nrel_id": "nrel-1", "id": "5a"}')]

    out = wizard._format_refs(refs, ids)

    assert expected == out



# -*- coding: utf-8 -*-
"""
Test
"""
import os
from elm import TEST_DATA_DIR
from elm.pdf import PDFtoTXT
import elm.pdf

os.environ["OPENAI_API_KEY"] = "dummy"

FP_PDF = os.path.join(TEST_DATA_DIR, 'GPT-4.pdf')
FP_TXT = os.path.join(TEST_DATA_DIR, 'gpt4.txt')

with open(FP_TXT, 'r', encoding='utf8') as f:
    TEXT = f.read()


class MockClass:
    """Dummy class to mock requests.post"""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def json(self):
        """Get json response"""
        content = self.kwargs['json']['messages'][1]['content']
        content = content[content.index('"""') + 3:]
        content = content[:content.index('"""')]
        response = {'choices': [{'message': {'content': content}}]}
        return response

    @classmethod
    def call(cls, **kwargs):
        """Mock for requests.post"""
        return cls(**kwargs)


def test_pdf_txt_clean(mocker):
    """Simple text on PDF text cleaning functionality.

    Note that LLM-based text cleaning is mocked here and not actually tested.
    """
    mocker.patch.object(elm.pdf.requests, "post", MockClass.call)
    pdf = PDFtoTXT(FP_PDF)
    pdf.clean_txt()

    missing = len(set(TEXT.split(' ')) - set(pdf.full.split(' ')))
    ntotal = len(TEXT.split(' '))

    assert (missing / ntotal) < 0.1
